/*
 *                   jfjji,.           H A M M E R H E A D  S H A R K W A R E
 *                  ,#######EL:                  http://www.hhsw.de
 *                   .,j######Kj       
 *                    ;W########G:     This  code  is  free software; you can
 *                   j#####G ,D#L:     redistribute it and/or modify it under
 *             .,;;,G#####D.           the  terms  of  the GNU General Public
 *           jE##########G             License   as  published  by  the  Free
 *             .,W######D.             Software  Foundation; either version 2
 *             :L#####K##G             of  the  License,  or (at your option)
 *           .f#####G;.t##:            any later version.
 *         ;L###ELi.    ,D,            
 *       iEEGWj                        This  code  is distributed in the hope
 *      tW;                            that  it  will  be useful, but without
 *     jf                              any warranty; without even the implied
 *    .Kt                              warranty of merchantability or fitness
 *    i:                               for  a particular purpose. See the GNU
 *    ;                                General   Public   License   for  more
 *    .                                details (http://www.gnu.org).
 */
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <iostream>
#include <string>

#include "Dock.hh"
#include "Slot.hh"

using namespace std;
using namespace bbdock;

/**
 * Initialize dock
 *
 * @param w - slot width in pixels (optional)
 * @param h - slot height in pixels (optional)
 */
Dock::Dock( int w, int h ) :
	slotwidth( w ),
	slotheight( h ),
	first( 0 )
{
	if(	!(display = XOpenDisplay( 0 )) )
		throw "Unable to open display";

	screen = DefaultScreen( display );
	screenwidth = DisplayWidth( display, screen );
	screenheight = DisplayHeight( display, screen );
	
	gc = XDefaultGC( display, screen );

	// determine bits per pixel of an screen-compatible XImage, does not 
	// need to be the same as the color-depth of the screen
	{
		XPixmapFormatValues	*pf;
		int					formats;

		bitsperpixel = DefaultDepth( display, screen );
		pf = XListPixmapFormats( display, &formats );

		if( formats )
		{
			int	format;

			for( format = formats;
			     format--; )
				if( pf[format].depth == bitsperpixel )
				{
					bitsperpixel = pf[format].bits_per_pixel;
					break; 
				}

			XFree( (char *) pf );
		}
	}
}

/**
 * Free resources  
 */
Dock::~Dock()
{
	XCloseDisplay( display );
}

/**
 * Add an icon to the dock
 *
 * @param icon - Icon object
 */
bool Dock::add( Icon *icon )
{
	if( !icon )
		return false;

	first = new Slot( *this, *icon, first );

	return true;
}

/**
 * Run the dock
 */
void Dock::run()
{
	if( !first )
		return;

	for( int xfd = ConnectionNumber( display ); ; )
	{
		if( !XPending( display ) )
		{
			Slot	*slot;

			for( slot = first;
				slot;
				slot = slot->getNext() )
				slot->tick();

			fd_set			rfds;
			struct timeval	tv;

			FD_ZERO( &rfds );
			FD_SET( xfd, &rfds );
			
			// suspend execution for a quarter second
			tv.tv_sec = 0;
			tv.tv_usec = 250000;

			select( xfd+1, &rfds, 0, 0, &tv );
			continue;
		}

		XEvent	event;

		bzero( &event, sizeof( event ) );
		XNextEvent( display, &event );

		Slot	*slot;

		for( slot = first;
			slot;
			slot = slot->getNext() )
			if( event.xany.window == slot->getWindow() )
				break;

		if( !slot )
		{
			if( event.type == ClientMessage )
				client( &event );

			continue;
		}

		switch( event.type )
		{
			case Expose:
				if( !event.xexpose.count )
					slot->draw();
				break;
			case VisibilityNotify:
				slot->setVisibility( event.xvisibility.state );
				break;
			case ConfigureNotify:
				slot->invalidate();
				slot->draw( true );
				break;
			case ButtonRelease:
				switch( event.xbutton.button )
				{
					case 1:
						slot->exec();
						break;
				}
				break;
		}
	}
}

/**
 * Send a message to a running instance of bbdock
 *
 * @param message - the message to send
 */
void Dock::send( const char *message, const char *data, int len )
{
	if( !message || !*message ||
		!data || !*data )
		return;

	Display	*display;
	int		screen;
	Window	root;

	if(	!(display = XOpenDisplay( 0 )) )
		throw "Unable to open display";

	screen = DefaultScreen( display );

	if( !(root = RootWindow( display, screen )) )
		throw "Can not access root window";

	Window			dummy;
	Window			*childs;
	unsigned int	n;

 	XQueryTree( display, root, &dummy, &dummy, &childs, &n );

	Atom	msg = XInternAtom( display, message, false );

	for( ; n--; )
	{
		XClassHint	xch;

		if( !XGetClassHint( display, childs[n], &xch ) )
			continue;

		if( !strcmp( xch.res_name, "bbdock" ) &&
			!strcmp( xch.res_class, "bbdock" ) )
		{
			XChangeProperty( display, childs[n], msg, msg, 8, 
				PropModeReplace, (unsigned char *) data, len );
		
			XEvent	event;

			event.xclient.type = ClientMessage;
			event.xclient.serial = 0;
			event.xclient.send_event = True;
			event.xclient.message_type = msg;
			event.xclient.window = childs[n];
			event.xclient.format = 32;
			event.xclient.data.l[0] = 0;
			event.xclient.data.l[1] = 0;
			event.xclient.data.l[2] = 0;
			event.xclient.data.l[3] = 0;
			event.xclient.data.l[4] = 0;

			XSendEvent( display, childs[n], True, NoEventMask, &event );

			// exit loop but not without freeing resources below
			n = 0;
		}

		XFree( xch.res_name );
		XFree( xch.res_class );
	}

	if( childs )
		XFree( childs );

	XCloseDisplay( display );
}

/**
 * Change icon of a command (remotely)
 *
 * @param cmd - command of icon to change
 * @param icon - new icon
 */
void Dock::changeIcon( const char *cmd, const char *icon )
{
	if( !cmd || !*cmd ||
		!icon || !*icon )
		return;

	string	data = cmd;

	data += ":";
	data += icon;

	Dock::send( "_BBDOCK_CHANGE_ICON_", data.c_str(), data.length() );
}

/**
 * Process a client message
 *
 * @param e - event
 */
void Dock::client( XEvent *e )
{
	char	*msg;

	if( !(msg = XGetAtomName( display, 
		e->xclient.message_type )) )
		return;

	Atom			returnedtype;
	int				format;
	unsigned long 	n;
	unsigned long	bytesafter;
	char			*prop;

	if( XGetWindowProperty( display, e->xany.window, e->xclient.message_type, 
		0, 1024, False, e->xclient.message_type, &returnedtype, &format, &n, 
		&bytesafter, (unsigned char **) &prop ) != Success )
		return;

	char	*cmd;
	char	*icon;
	
	if( (cmd = strtok( prop, ":" )) &&
		(icon = strtok( 0, "\n" )) )
	{
		Slot	*slot;

		for( slot = first;
			slot;
			slot = slot->getNext() )
			if( !strcasecmp( slot->getCommand(), cmd ) && 
				strcasecmp( slot->getIcon(), icon ) )
			{
				slot->setIcon( icon );
				break;
			}
	}

	XFree( prop );
	XDeleteProperty( display, e->xany.window, e->xclient.message_type );

	XFree( msg );
}
